window.amazon = window.amazon || {};
window.app = window.app || {};
if (!('willReappear' in window.app)) {
    window.app.willReappear = function() {};
}

(function (amazon) {
    "use strict";
    if (amazon.mash) {
        return;
    }
    // ------------------------------------------
    // MASH v1.1 API Module
    var Mash_v1_1 = {
        showChooser: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['buttonTitles', 'successCallback']);
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [
                    params.buttonTitles,
                    params.title,
                    params.cancelButtonTitle,
                    params.destructiveButtonTitle
                ];

                return execute1_1('showChooser', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },

        showAlert: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['message', 'buttonTitles', 'successCallback']);
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [
                    params.message,
                    params.buttonTitles,
                    params.title,
                    params.cancelButtonTitle
                ];

                return execute1_1('showAlert', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },

        openInExternalBrowser: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['url']);
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [params.url];
                return execute1_1('openInExternalBrowser', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },

        share: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['messagePlain', 'messagePlain160', 'messagePlain140', 'messageHTML', 'subject']);
                validateFunctions(params, ['successCallback', 'failCallback', 'shareServiceSelectedCallback']);

                var shareServiceSelectedCallbackId = amazon.mash.ios_util.storeCallback('shareServiceSelectedCallback', params.shareServiceSelectedCallback);

                params.args = [
                    params.messagePlain,
                    params.messagePlain160,
                    params.messagePlain140,
                    params.messageHTML,
                    params.subject,
                    params.imageURL,
                    params.url,
                    shareServiceSelectedCallbackId
                ];

                return execute1_1('share', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },

        getControl: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['controlId']);
                validateFunctions(params, ['successCallback', 'failCallback']);

                if (params.controlId === 'navButtonLeft' && !window.CDV_is_bottom_view_controller) {
                    throw 'invalidOperation';
                }

                var buttonId = buttonNamesToIds[params.controlId];
                var button = new ShellButton(buttonId);

                if (typeof params.successCallback === "function") {
                    params.successCallback();
                }

                return button;
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },

        cart: {
            didUpdate: function (params) {
                try {
                    params = params || {};
                    validateParameters(params, ['newCartQuantity']);
                    validateFunctions(params, ['successCallback', 'failCallback']);

                    params.args = [params.newCartQuantity];
                    return execute1_1('cartDidUpdate', params);
                } catch (error) {
                    handleError(error, params.failCallback);
                }
            }
        }
    };

    // ------------------------------------------
    // MASH v1.2 API Module
    var navigatePositions = [];
    var Mash_v1_2 = {
        showOfferListingPage: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['asin']);
                validateFunctions(params, ['successCallback', 'failCallback']);
                if ((params.listId && !params.listItemId) || (!params.listId && params.listItemId)) {
                    throw "listId and listItemId not supplied together";
                }
                params.args = [params.asin, params.listId, params.listItemId];
                return execute1_2('showOfferListingPage', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },
        addToWishList: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['asin']);
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [params.asin];
                return execute1_2('addToWishList', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },
        showImageGallery: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['title', 'imageUrls']);
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [
                    params.title,
                    params.startIndex,
                    params.imageUrls
                ];

                return execute1_2('showImageGallery', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },
        buyNowWithOneClick: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['asin', 'offerId', 'oneClickShippingSpeed']);
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [
                    params.asin,
                    params.offerId,
                    params.listId,
                    params.listItemId,
                    params.dealId,
                    params.clickStream,
                    params.oneClickShippingSpeed
                ];

                return execute1_2('buyNowWithOneClick', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },
        buyNow: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['asin', 'offerId']);
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [
                    params.asin,
                    params.offerId,
                    params.listId,
                    params.listItemId,
                    params.dealId,
                    params.clickStream
                ];

                return execute1_2('buyNow', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },
        navigate: function (params) {
            try {
                params = params || {};
                params.transitionEffect = params.transitionEffect || 'slide';
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [
                    params.url,
                    params.transitionEffect
                ];

                if (!params.url) {
                    navigatePositions.push({ x: window.scrollX, y: window.scrollY });
                }

                return execute1_2('navigate', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },
        goBack: function (params) {
            try {
                params = params || {};
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [];
                return execute1_2('goBack', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },
        showNotificationSettings: function (params) {
            try {
                params = params || {};
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [params.category];
                return execute1_2('showNotificationSettings', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },
        showProgressDialog: function (params) {
            try {
                params = params || {};
                validateParameters(params, ['dialogTitle']);
                validateFunctions(params, ['successCallback', 'failCallback', 'cancelCallback']);

                params.args = [params.dialogTitle];

                var successCallback = params.successCallback;
                var cancelCallback = params.cancelCallback;
                if (successCallback) {
                    params.successCallback = function(code) {
                        if (code === 'canceled') {
                            typeof cancelCallback === 'function' && cancelCallback();
                            return;
                        }

                        if (code === 'shown') {
                            successCallback();
                            return;
                        }
                    };
                }

                var commandId = execute1_2('showProgressDialog', params);
                if (params.cancelCallback) {
                    callbacks[commandId] = params.cancelCallback;
                }

                return commandId;
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },
        hideProgressDialog: function (params) {
            try {
                params = params || {};
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [];
                return execute1_2('hideProgressDialog', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        },
        showLoginDialog: function (params) {
            try {
                params = params || {};
                validateFunctions(params, ['successCallback', 'failCallback']);

                params.args = [];
                return execute1_2('showLoginDialog', params);
            } catch (error) {
                handleError(error, params.failCallback);
            }
        }
    };
 
    // ------------------------------------------
    // MASH Error
    var Error = function (code) {
        this.code = code;
    };
    
    Error.PERMISSION_DENIED = 1001;
    Error.TIMEOUT = 1002;
    Error.NOT_SUPPORTED = 1003;
    Error.OUT_OF_MEM = 1004;
    Error.UNKNOWN = 1005;
    Error.USER_CANCELLED = 1006;
    Error.USER_CANCELED = Error.USER_CANCELLED;  // Alternate spelling
    Error.INVALID_ARGUMENTS = 1007;
    Error.OPERATION_IN_PROGRESS = 1008;
 
    var BarcodeScanError = function (code) {
        this.code = code;
    };
 
    BarcodeScanError.CANNOT_OPEN_CAMERA = 1;

    var LaunchIntentURLError = function (code) {
        this.code = code;
    };
    
    LaunchIntentURLError.INTENT_NOT_FOUND = 1;

    var CardScannerError = function (code) {
        this.code = code;
    };

    CardScannerError.CANNOT_OPEN_CAMERA = 1;
    CardScannerError.INVALID_CARD_TYPE = 2;
    CardScannerError.NO_SUPPORTED_CARD_TYPES = 3;

    var LogoutError = function (code) {
        this.code = code;
    }
 
    LogoutError.NOT_LOGGED_IN = 1;
 
    // ------------------------------------------
    // MASH v1.3 API Module
    var Mash_v1_3 = {
        search: {
            registerForNotifications: function(params) {
                try {
                    params = params || {};
                    validateParameters1_3(params, ['notificationHandler']);
                    validateFunctions1_3(params, ['successCallback', 'failCallback']);
                    validateParameters1_3(params.notificationHandler, ['onPrefetch', 'onShowResults']);
                    validateFunctions1_3(params.notificationHandler, ['onPrefetch', 'onShowResults']);

                    amazon.mash.ios_util.search.listen('showresults', params.notificationHandler.onShowResults);
                    amazon.mash.ios_util.search.listen('prefetch', params.notificationHandler.onPrefetch);
                    params.successCallback && params.successCallback();
                    console.log('Search registered for notifications');
                } catch (error) {
                    handleError1_3(error, params.failCallback);
                }
            }
        },
        Error: Error
    };

    // ------------------------------------------
    // EventTarget Interface
    var Mash_EventTarget = (function () {
        var eventListeners = {};

        var getEventListenersForEvent = function (eventName) {
            if (!(eventName in eventListeners)) {
                eventListeners[eventName] = [];
            }
            return eventListeners[eventName];
        };

        var copyArray = function (arr) {
            return arr.slice(0);
        };

        var isHandlerRegisteredForEvent = function (eventName, handler) {
            var idx = getEventListenersForEvent(eventName).indexOf(handler);
            return idx !== -1;
        };

        var registerHandlerForEvent = function (eventName, handler) {
            if (isHandlerRegisteredForEvent(eventName, handler)) {
                return;
            }

            getEventListenersForEvent(eventName).push(handler);
        };

        return {
            addEventListener: function (eventName, handler) {
                // The event handler isn't registered until the successCallback is triggered, preventing
                // the API's use from within an Embedded Browser.
                var params = {};
                params.successCallback = function() { registerHandlerForEvent(eventName, handler); };
                execute1_4('addEventListener', params);
            },
            removeEventListener: function (eventName, handler) {
                var listeners = getEventListenersForEvent(eventName);
                var idx = listeners.indexOf(handler);
                if (idx !== -1) {
                    listeners.splice(idx, 1);
                }
            },
            dispatchEvent: function (eventObject) {
                // Asynchronously notify native + other webviews
                var params = {
                    args: [eventObject]
                };
                execute1_8('dispatchEvent', params);
            },
            ios_event_util: {
                dispatchLocally: function (eventObject) {
                    if (!('type' in eventObject)) {
                        throw 'invalid arguments';
                    }

                    var listeners = copyArray(getEventListenersForEvent(eventObject.type));
                    for (var i = 0; i < listeners.length; i++) {
                        var handler = listeners[i];
                        handler(eventObject);
                    }
                }
            }
        };
    }());

    // ------------------------------------------
    // MASH v1.4 API Module
    var Mash_v1_4 = {
        showContactPicker: function (params) {
            try {
                params = params || {};
                validateParameters1_3(params, ['fields']);
                validateFunctions1_3(params, ['successCallback', 'failCallback']);

                params.args = [params.fields];
                wrapFailCallbackWithHandleErrorCall(params, handleError1_4);

                return execute1_4('showContactPicker', params);
            } catch (error) {
                handleError1_3(error, params.failCallback);
            }
        },
        showOneClickSettings: function (params) {
            try {
                params = params || {};
                validateFunctions1_3(params, ['successCallback', 'failCallback']);

                wrapFailCallbackWithHandleErrorCall(params, handleError1_4);

                return execute1_4('showOneClickSettings', params);
            } catch (error) {
                handleError1_3(error, params.failCallback);
            }
        },
        launchIntentURL: function (params) {
            try {
                params = params || {};
                validateParameters1_3(params, ['url']);
                validateFunctions1_3(params, ['successCallback', 'failCallback']);

                params.args = [params.url];
                wrapFailCallbackWithHandleErrorCall(params, handleError1_4);

                return execute1_4('launchIntentURL', params);
            } catch (error) {
                handleError1_3(error, params.failCallback);
            }
        },
        canLaunchIntentURL: function (params) {
            try {
                params = params || {};
                validateParameters1_3(params, ['url', 'successCallback']);
                validateFunctions1_3(params, ['successCallback', 'failCallback']);

                params.args = [params.url];
                wrapFailCallbackWithHandleErrorCall(params, handleError1_4);
 
                return execute1_4('canLaunchIntentURL', params);
            } catch (error) {
                handleError1_3(error, params.failCallback);
            }
        },
        showDeliveryDestinationPicker: function (params) {
            try {
                params = params || {};
                validateFunctions1_3(params, ['successCallback', 'failCallback']);

                wrapFailCallbackWithHandleErrorCall(params, handleError1_4);

                return execute1_4('showDeliveryDestinationPicker', params);
            } catch (error) {
                handleError1_3(error, params.failCallback);
            }
        },
        showAdsPreferences: function(params) {
            try {
                params = params || {};
                validateFunctions(params, ['successCallback', 'failCallback']);

                wrapFailCallbackWithHandleErrorCall(params, handleError1_4);

                return execute1_4('showAdsPreferences', params);
            } catch (error) {
                handleError1_3(error, params.failCallback);
            }
        },
        requestFullscreen: function(params) {
            try {
                params = params || {};
                validateFunctions1_3(params, ['successCallback', 'failCallback']);

                wrapFailCallbackWithHandleErrorCall(params, handleError1_4);

                return execute1_4('requestFullscreen', params);
            } catch (error) {
                handleError1_3(error, params.failCallback);
            }
        },
        cancelFullscreen: function(params) {
            try {
                params = params || {};
                validateFunctions1_3(params, ['successCallback', 'failCallback']);

                wrapFailCallbackWithHandleErrorCall(params, handleError1_4);

                return execute1_4('cancelFullscreen', params);
            } catch (error) {
                handleError1_3(error, params.failCallback);
            }
        },
        showAmazonPoints: function(params) {
            try {
                params = params || {};
                validateFunctions1_3(params, ['successCallback', 'failCallback']);

                wrapFailCallbackWithHandleErrorCall(params, handleError1_4);

                return execute1_4('showAmazonPoints', params);
            } catch (error) {
                handleError1_3(error, params.failCallback);
            }
        },
        LaunchIntentURLError: LaunchIntentURLError
    };
 
    // ------------------------------------------
    // MASH v1.5 API Module
    var Mash_v1_5 = {
        showEmbeddedBrowser: function (params) {
            try {
                params = params || {};
                validateParameters1_3(params, ['targetURL', 'successCallback']);
                validateFunctions1_3(params, ['successCallback', 'failCallback']);
 
                params.args = [
                    params.targetURL,
                    params.method || "GET",
                    params.postParameters || {},
                    params.showAmazonLogo || false
                ];
                wrapFailCallbackWithHandleErrorCall(params, handleError1_5);
 
                return execute1_5('showEmbeddedBrowser', params);
            } catch (error) {
                handleError1_5(error, params.failCallback);
            }
        },
        exitEmbeddedBrowser: function (params) {
            try {
                params = params || {};
                validateParameters1_3(params, ['successCallback']);
                validateFunctions1_3(params, ['successCallback', 'failCallback']);

                params.args = [
                    params.targetURL,
                    params.method || "GET",
                    params.postParameters || {}
                ];
                wrapFailCallbackWithHandleErrorCall(params, handleError1_5);
 
                return executeEmbeddedBrowser('exitEmbeddedBrowser', params);
            } catch (error) {
                handleError1_5(error, params.failCallback);
            }
        },
        showBarcodeScanner: function (params) {
            try {
                params = params || {};
                validateParameters1_3(params, ['successCallback']);
                validateFunctions1_3(params, ['successCallback', 'failCallback']);
 
                wrapFailCallbackWithHandleErrorCall(params, handleError1_5);
 
                return execute1_5('showBarcodeScanner', params);
            } catch (error) {
                handleError1_5(error, params.failCallback);
            }
        },
        navMenu: {
            focusOnItem: function (params) {
                try {
                    params = params || {};
                    validateParameters1_3(params, ['id']);
                    validateFunctions1_3(params, ['successCallback', 'failCallback']);

                    params.args = [params.id];
 
                    wrapFailCallbackWithHandleErrorCall(params, handleError1_5);
 
                    return execute1_5('focusOnItem', params);
                } catch (error) {
                    handleError1_5(error, params.failCallback);
                }
            }
        },
        BarcodeScanError: BarcodeScanError
    };
 
    // ------------------------------------------
    // MASH v1.6 API Module
    var Mash_v1_6 = {
        showAdsBrowser: function (params) {
            try {
                params = params || {};
                validateParameters1_3(params, ['targetURL']);
                validateFunctions1_3(params, ['successCallback', 'failCallback']);

                params.args = [
                    params.targetURL
                ];
                wrapFailCallbackWithHandleErrorCall(params, handleError1_5);

                return execute1_6('showAdsBrowser', params);
            } catch (error) {
                handleError1_5(error, params.failCallback);
            }
        }
    };
 
    // ------------------------------------------
    // MASH v1.7 API Module
    var Mash_v1_7 = {
        showCardScanner: function (params) {
            try {
                params = params || {};
                validateParameters1_3(params, ['cardTypes', 'successCallback']);
                validateFunctions1_3(params, ['successCallback', 'failCallback']);
 
                params.args = [
                    params.cardTypes
                ];
                wrapFailCallbackWithHandleErrorCall(params, handleError1_7);
 
                return execute1_7('showCardScanner', params);
            } catch (error) {
                handleError1_7(error, params.failCallback);
            }
        },
        CardScannerError: CardScannerError,
 
        showShopByDepartment: function (params) {
            try {
                params = params || {};
                validateFunctions1_3(params, ['successCallback', 'failCallback']);
                wrapFailCallbackWithHandleErrorCall(params, handleError1_7);
                return execute1_7('showShopByDepartment', params);
            } catch (error) {
                handleError1_7(error, params.failCallback);
            }
        },
 
        logout: function (params) {
            try {
                params = params || {};
                validateFunctions1_3(params, ['successCallback', 'failCallback']);
                wrapFailCallbackWithHandleErrorCall(params, handleError1_7);
                return execute1_7('logout', params);
            } catch (error) {
                handleError1_7(error, params.failCallback);
            }
        },
        LogoutError: LogoutError, 
 
        showRecommendations: function (params) {
            try {
                params = params || {};
                validateFunctions1_3(params, ['successCallback', 'failCallback']);
                wrapFailCallbackWithHandleErrorCall(params, handleError1_7);
                return execute1_7('showRecommendations', params);
            } catch (error) {
                handleError1_7(error, params.failCallback);
            }
        },

        showYourAccount: function (params) {
            try {
                params = params || {};
                validateFunctions1_3(params, ['successCallback', 'failCallback']);
                wrapFailCallbackWithHandleErrorCall(params, handleError1_7);
                return execute1_7('showYourAccount', params);
            } catch (error) {
                handleError1_7(error, params.failCallback);
            }
        }
    }; 

    // ------------------------------------------
    // Shell Button (Private Type)
    var ShellButton = function (buttonId) {
        if (buttonId === undefined) {
            throw 'Element with the given ControlId could not be found.';
        }

        this.buttonId = buttonId;
    };

    ShellButton.prototype.show = function (params) {
        try {
            params = params || {};
            validateParameters(params, ['text', 'tapCallback']);
            validateFunctions(params, ['successCallback', 'failCallback', 'tapCallback']);

            callbacks['barButtonTapCallback' + this.buttonId] = params.tapCallback;

            params.args = [
                this.buttonId,
                params.text,
                validateBool(params.isEnabled, true),
                validateBool(params.isAnimated, true)
            ];

            return execute1_1('showBarButton', params);
        } catch (error) {
            handleError(error, params.failCallback);
        }
    };

    ShellButton.prototype.hide = function (params) {
        try {
            validateFunctions(params, ['successCallback', 'failCallback']);
            params.args = [
                this.buttonId,
                validateBool(params.isAnimated, true)
            ];
            return execute1_1('hideBarButton', params);
        } catch (error) {
            handleError(error, params.failCallback);
        }
    };

    ShellButton.prototype.setText = function (params) {
        try {
            params = params || {};
            validateParameters(params, ['text']);
            validateFunctions(params, ['successCallback', 'failCallback']);

            params.args = [
                this.buttonId,
                params.text
            ];

            return execute1_1('setButtonText', params);
        } catch (error) {
            handleError(error, params.failCallback);
        }
    };

    ShellButton.prototype.setEnabled = function (params) {
        try {
            params = params || {};
            validateParameters(params, ['isEnabled']);
            validateFunctions(params, ['successCallback', 'failCallback']);

            params.args = [
                this.buttonId,
                validateBool(params.isEnabled) // no default, because it's required
            ];

            return execute1_1('setButtonEnabledState', params);
        } catch (error) {
            handleError(error, params.failCallback);
        }
    };

    ShellButton.prototype.setTapCallback = function (params) {
        try {
            params = params || {};
            validateParameters(params, ['tapCallback']);
            validateFunctions(params, ['successCallback', 'failCallback', 'tapCallback']);

            callbacks['barButtonTapCallback' + this.buttonId] = params.tapCallback;

            if (typeof params.successCallback === "function") {
                params.successCallback();
            }
        } catch (error) {
            handleError(error, params.failCallback);
        }
    };

    // ------------------------------------------
    // document.title
    var documentTitleElement = document.getElementsByTagName('title')[0];
    var titleChangedHandler = function () {
        try {
            return execute1_4('updateDocumentTitle', {});
        } catch (error) {
            // No failCallback
        }
    };

    // ------------------------------------------
    // Private interface variables
    var buttonNamesToIds = {
        'navButtonLeft': 0,
        'navButtonRight': 1
    };

    var cordova = {}; // privately remember cordova
    var callbacks = {};
    var searchListeners = {
        'showresults': [],
        'prefetch': []
    };
    var backFenceCallbackIds = [];
    var hashChangeSteps = [];

    function popStateHandler(event) {
        if (!event.state) {
            console.log('ignoring popstate event, -- no state');
            return;
        }

        handleTransitionEvent();
    }

    function hashChangeHandler(event) {
        handleTransitionEvent();
    }

    function handleTransitionEvent() {
        window.removeEventListener('popstate', popStateHandler, false);
        window.removeEventListener('hashchange', hashChangeHandler, false);
        var steps = hashChangeSteps;
        hashChangeSteps = [];
        for (var i = 0; i < steps.length; i++) {
            typeof steps[i] === 'function' && steps[i]();
        }
    }

    var Mash_iOS_Util = {
        ios_util: {
            callCallback: function (callbackId, params, remove) {
                var callback = callbacks[callbackId];
                if (callback) {
                    callback.apply(callback, params);
                    if (remove) {
                        this.removeCallback(callbackId);
                    }
                }
            },
            storeCallback: function (prefix, callbackHandler) {
                if (typeof callbackHandler !== 'function') {
                    return null;
                }
                
                var callbackId = prefix;
                while (true) {
                    var suffix = Math.floor(Math.random()*1000000).toString();
                    if (!(callbackId + suffix in callbacks)) {
                        callbackId += suffix;
                        break;
                    }
                }
                callbacks[callbackId] = callbackHandler;
                return callbackId;
            },
            removeCallback: function (callbackId) {
                delete callbacks[callbackId];
            },
            popToBackFence: function (callbackId) {
                console.log('popToBackFence');
                hashChangeSteps.push(amazon.mash.ios_util.restoreIntraNavigationScrollPosition);
                if (callbackId) {
                    backFenceCallbackIds.push(callbackId);
                    hashChangeSteps.push(amazon.mash.ios_util.completeIntraPageBackTransition);
                }
                window.addEventListener('popstate', popStateHandler, false);
                window.addEventListener('hashchange', hashChangeHandler, false);
                // yield the UI thread
                setTimeout(function() { window.history.back(); }, 1);
            },
            restoreIntraNavigationScrollPosition: function () {
                console.log('restoreScrollPos');
                if (!navigatePositions.length) {
                    return true;
                }

                var scrollPos = navigatePositions.pop();
                if (typeof scrollPos.x !== 'undefined' && typeof scrollPos.y !== 'undefined') {
                    window.scrollTo(scrollPos.x, scrollPos.y);
                }
                return true;
            },
            completeIntraPageBackTransition: function() {
                console.log('completeIntraPageBackTransition');
                var callbackId = backFenceCallbackIds.pop();
                execute1_2('goBackCompleteIntraPageTransition', { args:[callbackId] });
            },
            search: {
                isListening: function(evt) {
                    evt = evt || 'showresults';
                    return searchListeners[evt].length > 0;
                },
                trigger: function(evt, data) {
                    var listeners = searchListeners[evt];
                    for (var i = 0; i < listeners.length; i++) {
                        listeners[i].apply(listeners[i], [data]);
                    }
                },
                listen: function(evt, func) {
                    if (!(evt in searchListeners)) {
                        throw 'internal error';
                    }

                    searchListeners[evt].push(func);
                }
            }
        }
    };

    var validateParameters = function (params, keys) {
        for (var i = 0; i < keys.length; ++i) {
            var key = keys[i];
            if (params[key] === undefined || params[key] === null) {
                throw 'Invalid or missing parameter: ' + key;
            }
        }
    };
    var validateParameters1_3 = function (params, keys) {
        try {
            validateParameters(params, keys);
        } catch (e) {
            throw new amazon.mash.Error(amazon.mash.Error.INVALID_ARGUMENTS);
        }
    };

    var validateFunctions = function (params, keys) {
        for (var i = 0; i < keys.length; ++i) {
            var key = keys[i];
            if (params[key] === undefined || params[key] === null) {
                continue;
            }
            if (typeof params[key] !== 'function') {
                throw 'Invalid function value for parameter: ' + key;
            }
        }
    };
    var validateFunctions1_3 = function (params, keys) {
        try {
            validateFunctions(params, keys);
        } catch (e) {
            throw new amazon.mash.Error(amazon.mash.Error.INVALID_ARGUMENTS);
        }
    };

    var validateBool = function (value, defaultValue) {
        if (value === undefined) {
            value = defaultValue;
        }

        if (value === 0) {
            value = false;
        } else if (value === 1) {
            value = true;
        }

        if (value !== true && value !== false) {
            throw 'Type mismatch';
        }

        return value;
    };
    var validateBool1_3 = function (value, defaultValue) {
        try {
            return validateBool(value, defaultValue);
        } catch (e) {
            throw new amazon.mash.Error(amazon.mash.Error.INVALID_ARGUMENTS);
        }
    };

    var execute = function (pluginName, methodName, params) {
        return cordova.exec(params.successCallback, params.failCallback, pluginName, methodName, params.args);
    };

    var execute1_1 = function (methodName, params) {
        return execute('AWAmazonShell1_1', methodName, params);
    };

    var execute1_2 = function (methodName, params) {
        return execute('AWAmazonShell1_2', methodName, params);
    };

    var execute1_3 = function (methodName, params) {
        return execute('AWAmazonShell1_3', methodName, params);
    };

    var execute1_4 = function (methodName, params) {
        return execute('AWAmazonShell1_4', methodName, params);
    };
 
    var execute1_5 = function (methodName, params) {
        return execute('AWAmazonShell1_5', methodName, params);
    };
 
    var execute1_6 = function (methodName, params) {
        return execute('AWAmazonShell1_6', methodName, params);
    };
 
    var execute1_7 = function (methodName, params) {
        return execute('AWAmazonShell1_7', methodName, params);
    };
 
    var execute1_8 = function (methodName, params) {
        return execute('AWAmazonShell1_8', methodName, params);
    };

    var executeEmbeddedBrowser = function (methodName, params) {
        return execute('AWAmazonShellEmbeddedBrowserPlugin', methodName, params);
    };

    var wrapFailCallbackWithHandleErrorCall = function(params, errorHandler) {
        var handler = params.failCallback;
        params.failCallback = function(e) { errorHandler(e, handler); };
    };

    var handleError = function (error, failCallback) {
        console.error('Error during MASH call: ' + error);
        if (typeof failCallback === "function") {
            failCallback.apply(failCallback, [error]);
        }
    };

    var handleError1_3 = function (error, failCallback) {
        try {
            if (error.constructor === amazon.mash.Error) {
                return handleError(error, failCallback);
            }
        } catch (e) {}

        if (typeof error === 'object' && 'code' in error && error.code in amazon.mash.Error) {
            handleError(new amazon.mash.Error(amazon.mash.Error[error.code]), failCallback);
        } else {
            console.error('Unknown error code during MASH call: ' + JSON.stringify(error));
            handleError(new amazon.mash.Error(amazon.mash.Error.UNKNOWN), failCallback);
        }
    };

    var handleError1_4 = function (error, failCallback) {
        try {
            if (error.constructor === amazon.mash.Error) {
                return handleError(error, failCallback);
            }
        } catch (e) {}

        if (typeof error === 'object' && 'code' in error && error.code in amazon.mash.LaunchIntentURLError) {
            handleError(new amazon.mash.Error(amazon.mash.LaunchIntentURLError[error.code]), failCallback);
        } else {
            handleError1_3(error, failCallback);
        }
    };
 
    var handleError1_5 = function (error, failCallback) {
        try {
            if (error.constructor === amazon.mash.Error) {
                return handleError(error, failCallback);
            }
        } catch (e) {}
 
        if (typeof error === 'object' && 'code' in error && error.code in amazon.mash.BarcodeScanError) {
            handleError(new amazon.mash.Error(amazon.mash.BarcodeScanError[error.code]), failCallback);
        } else {
            handleError1_4(error, failCallback);
        }
    };

    var handleError1_7 = function (error, failCallback) {
        try {
            if (error.constructor === amazon.mash.Error) {
                return handleError(error, failCallback);
            }
        } catch (e) {}
 
        if (typeof error === 'object' && 'code' in error && error.code in amazon.mash.CardScannerError) {
            handleError(new amazon.mash.Error(amazon.mash.CardScannerError[error.code]), failCallback);
        } else if (typeof error === 'object' && 'code' in error && error.code in amazon.mash.LogoutError) {
            handleError(new amazon.mash.Error(amazon.mash.LogoutError[error.code]), failCallback);
        } else {
            handleError1_5(error, failCallback);
        }
    };

    // combineModules can be used to pass successive module versions and get a final published module
    // i.e. var mash_v1x = combine_modules(mash_v11, mash_v12, ...)
    // i.e. var mash_v2x = combine_modules(mash_v20)
    var combineModules = function () {
        var obj = {},
            numArgs = arguments.length;
        for (var i = 0; i < numArgs; i++) {
            var module = arguments[i];
            for (var prop in module) {
                if (module.hasOwnProperty(prop)) {
                    obj[prop] = module[prop];
                }
            }
        }

        return obj;
    };

    // Cordova APIs that remain accessible in window.cordova (needed by the app)
    var globalCordovaWhitelist = {
            'require': 'require',
            'callbackSuccess' : 'callbackSuccess',
            'callbackError' : 'callbackError',
            'callbackFromNative' : 'callbackFromNative',
            'fireDocumentEvent' : 'fireDocumentEvent',
            'fireWindowEvent': 'fireWindowEvent'
        };

    // Cordova APIs that remain accessible in amazon.cordova (needed by feature pages)
    var amazonCordovaWhitelist = {
        // keypath in amazon.cordva : keypath in window.cordova
            'notification.vibrate' : 'apis.navigator.notification.vibrate'
        };

    // Cordova APIs that are accessible in navigator.*
    var globalNavigatorWhitelist = {
        'geolocation': 'apis.navigator.geolocation'
    };

    // Utility function to force an object hierarchy to exist (e.g. forceKeyPathExistance(fooObject, 'foo.bar.baz.bot');)
    var forceKeyPathExistance = function (object, keypath) {
        var keys = keypath.split('.');
        var target = object;
        var prevtarget;
        for (var i = 0; i < keys.length; i++) {
            if (!(keys[i] in target)) {
                target[keys[i]] = {};
            }

            prevtarget = target;
            target = target[keys[i]];
        }

        return { container: prevtarget, lastKey: keys[keys.length - 1] };
    };

    // Utility function to get an object from an object hierarchy (e.g. getObjectFromKeyPath(fooObject, 'foo.bar.baz.bot');)
    var getObjectFromKeyPath = function (object, keypath) {
        var keys = keypath.split('.');
        var target = object;
        for (var i = 0; i < keys.length; i++) {
            if (!(keys[i] in target)) {
                return undefined;
            }

            target = target[keys[i]];
        }
        return target;
    };

    // Utility function to return a javascript object reduced by a whitelist
    var applyWhitelist = function (baseObject, whitelist) {
        var wrapper = {};
        for (var wrapperKey in whitelist) {
            if (whitelist.hasOwnProperty(wrapperKey)) {
                var whitelistedObject = getObjectFromKeyPath(baseObject, whitelist[wrapperKey]);
                if (typeof whitelistedObject !== 'undefined') {
                    var insertLocation = forceKeyPathExistance(wrapper, wrapperKey);
                    insertLocation.container[insertLocation.lastKey] = whitelistedObject;
                }
            }
        }
        return wrapper;
    };

    // Utility function to merge objects from a whitelist into an existing object
    var applyWhitelistInto = function(baseObject, whitelist, target) {
        var whitelistedObject = applyWhitelist(baseObject, whitelist);
        for (var p in whitelistedObject) {
            if (whitelistedObject.hasOwnProperty(p)) {
                target[p] = whitelistedObject[p];
            }
        }
    };

    var bootstrapComplete = false;
    var onDeviceReady = function () {
        if (bootstrapComplete) {
            console.log('bootstrap already completed');
            return;
        }

        if (!window.cordova) {
            console.log('Cordova is not included on this page. You must include Cordova to use the MASH library.');
            return;
        }

        // Privately remember cordova
        cordova = window.cordova;
        // Disable global access to cordova except calls which the view controller needs
        window.cordova = applyWhitelist(cordova, globalCordovaWhitelist);
        // Expose whitelisted cordova objects as aliases in amazon.cordova
        amazon.cordova = applyWhitelist(cordova, amazonCordovaWhitelist);
        // Expose whitelisted cordova.navigator objects into navigator
        applyWhitelistInto(cordova, globalNavigatorWhitelist, navigator);
        // Expose MASH

        var mashComposition = [
                Mash_v1_1,
                Mash_v1_2,
                Mash_v1_3,
                Mash_v1_4,
                Mash_v1_5,
                Mash_v1_6,
                Mash_v1_7,
                Mash_EventTarget,
                Mash_iOS_Util
            ];

        amazon.mash = combineModules.apply(combineModules, mashComposition);

        // Listen for Cordova's pause and resume events and dispatch them as mash appPause and appResume events
        document.addEventListener('pause', function() {
            amazon.mash.ios_event_util.dispatchLocally({type:'appPause', detail:undefined});
        }, false);
        document.addEventListener('resume', function() {
            amazon.mash.ios_event_util.dispatchLocally({type:'appResume', detail:undefined});
        }, false);

        // title change listener
        if (documentTitleElement) {
            documentTitleElement.addEventListener('DOMSubtreeModified', titleChangedHandler, false);
        }
 
        // Bootstrap now complete. Mark as such and call back to application JS
        bootstrapComplete = true;
        var amznJQ = window.amznJQ;
        var app = window.app;
        if (amznJQ && typeof amznJQ.declareAvailable === "function") {
            amznJQ.declareAvailable('mash');
        }

        if (app && typeof app.ready === "function") {
            app.ready();
        }
        // aui-compatibility layer
        if (typeof P === "object" && typeof P.register === "function") {
            P.register("mash", function(){ return amazon.mash; });
        }
    };

    var bootstrapCordova;
    if (window.cordova) {
        bootstrapCordova = function () {
            var channel = window.cordova.require('cordova/channel');
            channel.onDeviceReady.subscribe(onDeviceReady);
        };
    } else {
        bootstrapCordova = function () {
            var head = document.head || document.getElementsByTagName('head')[0] || document.documentElement || document.getElementsByTagName('html')[0];
            var script = document.createElement('script');
            script.async = 'async';
            script.onload = function () {
                var channel = window.cordova.require('cordova/channel');
                channel.onDeviceReady.subscribe(onDeviceReady);
            };
            script.src = 'https://load/cordova/MASHCordova-iOS.js';
            head.appendChild(script);
        };
    }

    bootstrapCordova();
}(window.amazon));
